//*************************************************************************
// JANUS is a simple, robust, open standard signalling method for         *
// underwater communications. See <http://www.januswiki.org> for details. *
//*************************************************************************
// Example software implementations provided by STO CMRE are subject to   *
// Copyright (C) 2008-2018 STO Centre for Maritime Research and           *
// Experimentation (CMRE)                                                 *
//                                                                        *
// This is free software: you can redistribute it and/or modify it        *
// under the terms of the GNU General Public License version 3 as         *
// published by the Free Software Foundation.                             *
//                                                                        *
// This program is distributed in the hope that it will be useful, but    *
// WITHOUT ANY WARRANTY; without even the implied warranty of FITNESS     *
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for       *
// more details.                                                          *
//                                                                        *
// You should have received a copy of the GNU General Public License      *
// along with this program. If not, see <http://www.gnu.org/licenses/>.   *
//*************************************************************************
// Author: Ricardo Martins                                                *
//*************************************************************************

#include <janus/janus.h>

// JANUS headers.
#include <janus/types.h>

#if defined(JANUS_CPU_WORD_8BIT)

static const janus_uint8_t c_prime_table[] =
{
  0xac, 0x28, 0x8a, 0xa0, 0x20, 0x8a, 0x20, 0x28, 0x88, 0x82, 0x08, 0x02,
  0xa2, 0x28, 0x02, 0x80, 0x08, 0x0a, 0xa0, 0x20, 0x88, 0x20, 0x28, 0x80,
  0xa2, 0x00, 0x08, 0x80, 0x28, 0x82, 0x02, 0x08, 0x82, 0xa0, 0x20, 0x0a,
  0x20, 0x00, 0x88, 0x22, 0x00, 0x08, 0x02, 0x28, 0x82, 0x80, 0x20, 0x88,
  0x20, 0x20, 0x02, 0x02, 0x28, 0x80, 0x82, 0x08, 0x02, 0xa2, 0x08, 0x80,
  0x80, 0x08, 0x88, 0x20, 0x00, 0x0a, 0x00, 0x20, 0x08, 0x20, 0x08, 0x0a,
  0x02, 0x08, 0x82, 0x82, 0x20, 0x0a, 0x80, 0x00, 0x8a, 0x20, 0x28, 0x00,
  0x22, 0x08, 0x08, 0x20, 0x20, 0x80, 0x80, 0x20, 0x88, 0x80, 0x20, 0x02,
  0x22, 0x00, 0x08, 0x20, 0x00, 0x0a, 0xa0, 0x28, 0x80, 0x00, 0x20, 0x8a,
  0x00, 0x20, 0x8a, 0x00, 0x00, 0x88, 0x80, 0x00, 0x02, 0x22, 0x08, 0x02,
  0x80, 0x08, 0x82, 0x80, 0x20, 0x00, 0x22, 0x28
};

int
janus_utils_primes_is_prime(unsigned nr)
{
 return (c_prime_table[nr >> 3] & (1 << (nr & 7))) != 0;
}

#elif defined(JANUS_CPU_WORD_16BIT)

static const janus_uint16_t c_prime_table[] =
{
  0x28ac, 0xa08a, 0x8a20, 0x2820, 0x8288, 0x0208, 0x28a2, 0x8002, 0x0a08,
  0x20a0, 0x2088, 0x8028, 0x00a2, 0x8008, 0x8228, 0x0802, 0xa082, 0x0a20,
  0x0020, 0x2288, 0x0800, 0x2802, 0x8082, 0x8820, 0x2020, 0x0202, 0x8028,
  0x0882, 0xa202, 0x8008, 0x0880, 0x2088, 0x0a00, 0x2000, 0x2008, 0x0a08,
  0x0802, 0x8282, 0x0a20, 0x0080, 0x208a, 0x0028, 0x0822, 0x2008, 0x8020,
  0x2080, 0x8088, 0x0220, 0x0022, 0x2008, 0x0a00, 0x28a0, 0x0080, 0x8a20,
  0x2000, 0x008a, 0x8800, 0x0080, 0x2202, 0x0208, 0x0880, 0x8082, 0x0020,
  0x2822
};

int
janus_utils_primes_is_prime(unsigned nr)
{
 return (c_prime_table[nr >> 4] & (1 << (nr & 15))) != 0;
}

#else

static const janus_uint32_t c_prime_table[] =
{
  0xa08a28ac, 0x28208a20, 0x02088288, 0x800228a2, 0x20a00a08, 0x80282088,
  0x800800a2, 0x08028228, 0x0a20a082, 0x22880020, 0x28020800, 0x88208082,
  0x02022020, 0x08828028, 0x8008a202, 0x20880880, 0x20000a00, 0x0a082008,
  0x82820802, 0x00800a20, 0x0028208a, 0x20080822, 0x20808020, 0x02208088,
  0x20080022, 0x28a00a00, 0x8a200080, 0x008a2000, 0x00808800, 0x02082202,
  0x80820880, 0x28220020
};

int
janus_utils_primes_is_prime(unsigned nr)
{
 return (c_prime_table[nr >> 5] & (1 << (nr & 31))) != 0;
}

#endif

unsigned
janus_utils_primes_get_previous(unsigned nr)
{
  while (nr)
  {
    if (janus_utils_primes_is_prime(nr))
      return nr;

    --nr;
  }

  return 0;
}
